%x COMMENT
%x BRACKETS

%{
#include <stdlib.h>

#include "general.h"
#include "error.h"
#include "quad.h"
#include "parser.h"

#define T_eof 0
int brackets=0;
%}

L [A-Za-z]
D [0-9]
H [0-9A-Fa-f]
C (([^\n\\\'\"])|(\\("n"|"t"|"r"|"0"|\\|"'"|\"|("x"{H}{2}))))

%%

<<EOF>>     { return T_eof; }

"unit"      { return T_unit; }
"endu"      { return T_endu; }
":="        { return T_assign; }
"array"     { return T_array; }
"<>"        { return T_ne; }
"<="        { return T_le; }
">="        { return T_ge; }
"ifb"       { return T_ifb; }
"jump"      { return T_jump; }
"label"     { return T_label; }
"jumpl"     { return T_jumpl; }
"call"      { return T_call; }
"par"       { return T_par; }
"ret"       { return T_ret; }
"dim"       { return T_dim; }
"=="        { return T_eqeq; }
"!="        { return T_nxeq; }
       
"true"      { return T_true; }
"false"     { return T_false; }

"+"         { return '+'; }
"-"         { return '-'; }
"*"         { return '*'; }
"/"         { return '/'; }
"%"         { return '%'; }
"="         { return '='; }
"<"         { return '<'; }
">"         { return '>'; }
":"         { return ':'; }
";"         { return ';'; }
","         { return ','; }
"["         { return '['; }
"]"         { return ']'; }
"{"         { return '{'; }
"}"         { return '}'; }

"$$"        { return T_result; }

"V"         { return 'V'; }
"R"         { return 'R'; }
"RET"       { return T_RET; }

{D}+        { yylval.n = atoi(yytext); return T_cint; }

{D}+"."{D}+([Ee]("+"|"-")?{D}+)? { yylval.f = atof(yytext); return T_cfloat; }

"'"{C}"'"   { yylval.c = strdup(yytext); return T_character; }

\"{C}*\"    { yylval.s = strdup(yytext); return T_string; }

[ \t]+      { /*nothing*/ }
\n          { linecount++; return T_nl; }

"obj"                         { BEGIN(BRACKETS); return T_obj; }
"fun"                         { BEGIN(BRACKETS); return T_fun; }
"temp"                        { BEGIN(BRACKETS); return T_temp; }
<BRACKETS>[ \t]+              { /* nothing */ }
<BRACKETS>\n                  { linecount++; }
<BRACKETS>";"                 { return ';'; }
<BRACKETS>"array"             { return T_array; }
<BRACKETS>"ref"               { return T_ref; }
<BRACKETS>"of"                { return T_of; }
<BRACKETS>"unit"              { return T_unit; }
<BRACKETS>"int"               { return T_int; }
<BRACKETS>"char"              { return T_char; }
<BRACKETS>"bool"              { return T_bool; }
<BRACKETS>"float"             { return T_float; }
<BRACKETS>"->"                { return T_arrow; }
<BRACKETS>"+"                 { return '+'; }
<BRACKETS>"-"                 { return '-'; }
<BRACKETS>"*"                 { return '*'; }
<BRACKETS>"("                 { return '('; }
<BRACKETS>")"                 { return ')'; }
<BRACKETS>"["                 { ++brackets; return '['; }
<BRACKETS>","                 { return ','; }
<BRACKETS>({L}|_)({L}|{D}|_)* { yylval.obj.id = id_make(yytext); return T_id; }
<BRACKETS>-?{D}+              { yylval.n = atoi(yytext); return T_cint; }
<BRACKETS>"$"{D}+             { yylval.n = atoi(yytext+1); return T_tempid; }
<BRACKETS>"]"                 { if (--brackets == 0) BEGIN(INITIAL);
                                return ']'; }
<BRACKETS>.    { error("illegal token in object, function or temporary: "
                       "'%s'", yytext); }

"/*"              { BEGIN(COMMENT); }
<COMMENT>"*/"     { BEGIN(INITIAL); }
<COMMENT>\n       { linecount++; }
<COMMENT>"*"      { /*nothing*/ }
<COMMENT>[^*\n]+  { /*nothing*/ }

.           { error("illegal token: '%s'", yytext); }

%%
